#include "runner.hpp"

#include <algorithm>
#include <array>
#include <cerrno>
#include <fcntl.h>
#include <poll.h>
#include <spawn.h>
#include <sys/timerfd.h>
#include <sys/wait.h>
#include <system_error>
#include <unistd.h>

// Define a helper to throw a system error if a syscall fails
static auto ensure = [](int res) -> void {
    if (res == -1) {
        throw std::system_error{errno, std::generic_category()};
    }
};

namespace sk {
runner::runner(runner_backend backend, const runner_config &config) : backend{backend}, timeout{static_cast<int>(config.timeout)} {}

run_result runner::run_blocking(const program &program) {
    // Open file descriptors ahead of time
    int in_pipe[2];
    int out_pipe[2];
    int err_pipe[2];
    if (pipe2(in_pipe, O_CLOEXEC) == -1 || pipe2(out_pipe, O_CLOEXEC) == -1 || pipe2(err_pipe, O_CLOEXEC) == -1) {
        throw std::system_error{errno, std::generic_category()};
    }

    // Create a timer that will be polled when the program runs too long
    int timerfd = timerfd_create(CLOCK_MONOTONIC, TFD_CLOEXEC);
    if (timerfd == -1) {
        throw std::system_error{errno, std::generic_category()};
    }
    itimerspec timer{};
    timer.it_value.tv_sec = timeout;
    if (timerfd_settime(timerfd, 0, &timer, nullptr) == -1) {
        throw std::system_error{errno, std::generic_category()};
    }

    // Avoid blocking on stdin, and being interrupted if the pipe is closed prematurely
    int flags = fcntl(in_pipe[1], F_GETFL, 0);
    fcntl(in_pipe[1], F_SETFL, flags | O_NONBLOCK);

    posix_spawn_file_actions_t actions;
    posix_spawn_file_actions_init(&actions);
    posix_spawn_file_actions_addclose(&actions, timerfd);
    posix_spawn_file_actions_adddup2(&actions, in_pipe[0], STDIN_FILENO);
    posix_spawn_file_actions_adddup2(&actions, out_pipe[1], STDOUT_FILENO);
    posix_spawn_file_actions_adddup2(&actions, err_pipe[1], STDERR_FILENO);
    const char *const docker_args[] = {"docker", "run", "--rm", "-i", "--name", program.name.c_str(), "--pull=never", "--cap-drop=ALL", "--network=none", "--memory=64m", "--memory-swap=64m", "--pids-limit=128", program.image.c_str(), nullptr};
    const char *const bwrap_args[] = {"bwrap", "--ro-bind", "/usr", "/usr", "--dir", "/tmp", "--dir", "/var", "--proc", "/proc", "--dev", "/dev", "--symlink", "usr/lib", "/lib", "--symlink", "usr/lib64", "/lib64", "--symlink", "usr/bin", "/bin", "--symlink", "usr/sbin", "/sbin", "--unshare-all", "--die-with-parent", "/bin/sh", nullptr};
    const char *const *args = docker_args;
    if (backend == runner_backend::BubbleWrap) {
        args = bwrap_args;
    }
    pid_t pid;
    bool killed = false;
    int exit_code;
    if (posix_spawnp(&pid, args[0], &actions, nullptr, const_cast<char *const *>(args), nullptr) != 0) {
        throw std::system_error{errno, std::generic_category()};
    }

    close(in_pipe[0]);
    close(out_pipe[1]);
    close(err_pipe[1]);

    // Register the job as active
    {
        std::lock_guard<std::mutex> guard(active_jobs_mutex);
        active_jobs.push_back(active_job{program.name, pid});
    }

    size_t len = program.code.size();
    size_t window = 0;
    const char *data = program.code.data();

    std::array<char, 1024> buffer{};
    std::string out;
    std::string err;
    std::array<pollfd, 4> plist = {pollfd{in_pipe[1], POLLOUT | POLLHUP, 0}, pollfd{out_pipe[0], POLLIN, 0}, pollfd{err_pipe[0], POLLIN, 0}, pollfd{timerfd, POLLIN, 0}};
    pollfd *pfds = plist.data();
    nfds_t nfds = plist.size();

    int res;
    while ((res = poll(pfds, nfds, -1)) > 0 || (res == -1 && errno == EINTR)) {
        if (res == -1) {
            // Interrupted by a signal, retry
            continue;
        }

        nfds_t polled_fds = 0;
        if (nfds == plist.size()) {
            // Poll stdin when we still have data to write
            if (pfds[0].revents & (POLLHUP | POLLERR)) {
                // Closing input prematurely
                close(in_pipe[1]);
                --nfds;
                ++pfds;
                ++polled_fds;
            } else if (pfds[0].revents & POLLOUT) {
                if (window >= len) {
                    // End input
                    close(in_pipe[1]);
                    --nfds;
                    ++pfds;
                } else {
                    // Write the current data window
                    size_t l = std::min(len - window, buffer.size());
                    write(pfds[0].fd, data + window, l);
                    window += l;
                }
                ++polled_fds;
            }
        }

        // Poll stdout, stderr and the timer
        for (nfds_t i = nfds - 3; i < nfds; ++i) {
            if (pfds[i].revents & POLLIN) {
                ssize_t bytes_read = read(pfds[i].fd, buffer.data(), buffer.size());
                if (bytes_read == -1) {
                    throw std::system_error{errno, std::generic_category()};
                }
                if (pfds[i].fd == out_pipe[0]) {
                    out.append(buffer.data(), bytes_read);
                } else if (pfds[i].fd == timerfd) {
                    std::lock_guard<std::mutex> guard(active_jobs_mutex);
                    auto it = std::find_if(active_jobs.begin(), active_jobs.end(), [pid](const active_job &job) { return job.pid == pid; });
                    if (it != active_jobs.end()) {
                        exit(*it);
                        active_jobs.erase(it);
                    }
                    killed = true;
                } else {
                    err.append(buffer.data(), bytes_read);
                }
                ++polled_fds;
            }
        }

        // If nothing was polled, we're done
        if (polled_fds == 0) {
            break;
        }
    }

    waitpid(pid, &exit_code, 0);
    close(out_pipe[0]);
    close(err_pipe[0]);
    close(timerfd);

    // Remove the job from the active list
    {
        std::lock_guard<std::mutex> guard(active_jobs_mutex);
        auto it = std::find_if(active_jobs.begin(), active_jobs.end(), [pid](const active_job &job) { return job.pid == pid; });
        if (it != active_jobs.end()) {
            active_jobs.erase(it);
        }
    }

    posix_spawn_file_actions_destroy(&actions);
    return run_result{out, err, killed ? 124 : exit_code};
}

bool runner::kill_active(const std::string &jobId) {
    std::lock_guard<std::mutex> guard(active_jobs_mutex);
    auto it = std::find_if(active_jobs.begin(), active_jobs.end(), [&jobId](const active_job &job) { return job.job_id == jobId; });
    if (it != active_jobs.end()) {
        exit(*it);
        active_jobs.erase(it);
        return true;
    }
    return false;
}

void runner::exit_active_jobs() {
    std::lock_guard<std::mutex> guard(active_jobs_mutex);
    for (const auto &job : active_jobs) {
        exit(job);
    }
    active_jobs.clear();
}

void runner::exit(const active_job &job) {
    if (backend == runner_backend::Docker) {
        const char *const kill_args[] = {"docker", "kill", job.job_id.c_str(), nullptr};
        pid_t kill_pid;
        ensure(posix_spawnp(&kill_pid, kill_args[0], nullptr, nullptr, const_cast<char *const *>(kill_args), nullptr));
    } else {
        ensure(kill(job.pid, SIGINT));
    }
}
}
